@@ -61,6 +61,14 @@ type Config struct {
 	// Maidenhead grid square (e.g. JP20qe).
 	Locator string `json:"locator"`
 
+	// Message size limit (in bytes) for automatic download of pending messages.
+	//
+	// When receiving messages larger than this limit, the user will be prompted
+	// with the option to defer the download to a later session.
+	//
+	// Negative value means no limit.
+	AutoDownloadSizeLimit int `json:"auto_download_size_limit"`
+
 	// List of service codes for rmslist (defaults to PUBLIC)
 	ServiceCodes []string `json:"service_codes"`
 
@@ -101,6 +109,9 @@ type Config struct {
 	// See GPSdConfig.
 	GPSd GPSdConfig `json:"gpsd"`
 
+	// See PredictionConfig.
+	Prediction PredictionConfig `json:"prediction,omitzero"`
+
 	// Legacy support for old config files only. This field is deprecated!
 	// Please use "Addr" field in GPSd config struct (GPSd.Addr)
 	GPSdAddrLegacy string `json:"gpsd_addr,omitempty"`
@@ -123,6 +134,63 @@ type Config struct {
 	VersionReportingDisabled bool `json:"version_reporting_disabled"`
 }
 
+func (c *Config) UnmarshalJSON(b []byte) error {
+	type aliased Config
+	tmp := struct {
+		aliased
+		// Default to -1 if omitted (legacy configs)
+		AutoDownloadSizeLimit *int `json:"auto_download_size_limit"`
+	}{}
+	if err := json.Unmarshal(b, &tmp); err != nil {
+		return err
+	}
+	if ptr := tmp.AutoDownloadSizeLimit; ptr == nil {
+		tmp.aliased.AutoDownloadSizeLimit = -1
+	} else {
+		tmp.aliased.AutoDownloadSizeLimit = *tmp.AutoDownloadSizeLimit
+	}
+	*c = Config(tmp.aliased)
+	return nil
+}
+
+type PredictionConfig struct {
+	// engine sets the HF propagation prediction engine to be used.
+	// Valid options are:
+	//   - empty string (autodetect)
+	//   - voacap
+	//   - voacap-api
+	//   - disabled
+	// If empty, autodetection is attempted.
+	Engine PredictionEngine `json:"engine"`
+
+	// See VOACAPConfig.
+	VOACAP VOACAPConfig `json:"voacap,omitzero"`
+
+	// See VOACAPAPIConfig.
+	VOACAPAPI VOACAPAPIConfig `json:"voacap_api,omitzero"`
+}
+
+func (p PredictionConfig) IsZero() bool { return p == (PredictionConfig{}) }
+
+type VOACAPConfig struct {
+	// Executable overrides the default executable for VOACAP.
+	// Default is c:\itshfbc\bin_win\voacapw.exe on Windows and voacapl for other platforms.
+	Executable string `json:"executable"`
+	// DataDir overrides the default data directory for VOACAP.
+	// Default is c:\itshfbc on Windows and $HOME/itshfbc for other platforms.
+	DataDir string `json:"data_dir"`
+}
+
+func (v VOACAPConfig) IsZero() bool { return v == (VOACAPConfig{}) }
+
+type VOACAPAPIConfig struct {
+	// BaseURL is the base URL for the VOACAP API.
+	// E.g. http://localhost:3000
+	BaseURL string `json:"base_url"`
+}
+
+func (v VOACAPAPIConfig) IsZero() bool { return v == (VOACAPAPIConfig{}) }
+
 type HamlibConfig struct {
 	// The network type ("serial" or "tcp"). Use 'tcp' for rigctld (default).
 	//
@@ -146,6 +214,11 @@ type ArdopConfig struct {
 	// Default/listen ARQ bandwidth (200/500/1000/2000 MAX/FORCED).
 	ARQBandwidth ardop.Bandwidth `json:"arq_bandwidth"`
 
+	// Number of connect frames to transmit when dialing.
+	//
+	// Some stations may require mutliple attempts before responding.
+	ConnectRequests int `json:"connect_requests"`
+
 	// (optional) Reference name to the Hamlib rig to control frequency and ptt.
 	Rig string `json:"rig"`
 
@@ -164,7 +237,7 @@ type VaraConfig struct {
 	Addr string `json:"addr"`
 
 	// Default/listen bandwidth (HF: 500/2300/2750 Hz).
-	Bandwidth int `json:"bandwidth"`
+	Bandwidth int `json:"bandwidth,omitempty"`
 
 	// (optional) Reference name to the Hamlib rig to control frequency and ptt.
 	Rig string `json:"rig"`
@@ -309,14 +382,21 @@ type GPSdConfig struct {
 	// Use server time instead of timestamp provided by GPSd (e.g for older GPS device with week roll-over issue).
 	UseServerTime bool `json:"use_server_time"`
 
+	// Automatically update the locator field in-memory by polling GPSd every hour.
+	//
+	// Note: This only updates the locator in-memory. The config file is not modified.
+	// On startup, the config's locator value will be used until the first position is received from GPSd.
+	UpdateLocator bool `json:"update_locator"`
+
 	// Address and port of GPSd server (e.g. localhost:2947).
 	Addr string `json:"addr"`
 }
 
 var DefaultConfig = Config{
-	MOTD:         []string{"Open source Winlink client - getpat.io"},
-	AuxAddrs:     []AuxAddr{},
-	ServiceCodes: []string{"PUBLIC"},
+	MOTD:                  []string{"Open source Winlink client - getpat.io"},
+	AuxAddrs:              []AuxAddr{},
+	ServiceCodes:          []string{"PUBLIC"},
+	AutoDownloadSizeLimit: -1,
 	ConnectAliases: map[string]string{
 		"telnet": "telnet://{mycall}:CMSTelnet@cms.winlink.org:8772/wl2k",
 	},
@@ -344,9 +424,10 @@ var DefaultConfig = Config{
 		RadioPort: 0,
 	},
 	Ardop: ArdopConfig{
-		Addr:         "localhost:8515",
-		ARQBandwidth: ardop.Bandwidth500Max,
-		CWID:         true,
+		Addr:            "localhost:8515",
+		ARQBandwidth:    ardop.Bandwidth500Max,
+		ConnectRequests: ardop.DefaultConnectRequests,
+		CWID:            true,
 	},
 	Pactor: PactorConfig{
 		Path:     "/dev/ttyUSB0",
@@ -367,6 +448,7 @@ var DefaultConfig = Config{
 		EnableHTTP:    false, // Default to false to help protect privacy of unknowing users (see github.com//issues/146)
 		AllowForms:    false, // Default to false to help protect location privacy of unknowing users
 		UseServerTime: false,
+		UpdateLocator: false,
 		Addr:          "localhost:2947", // Default listen address for GPSd
 	},
 	GPSdAddrLegacy: "",
