@@ -7,6 +7,7 @@
 package forms
 
 import (
+	"bytes"
 	"context"
 	"encoding/json"
 	"encoding/xml"
@@ -14,7 +15,6 @@ import (
 	"fmt"
 	"io"
 	"io/fs"
-	"io/ioutil"
 	"log"
 	"net/http"
 	"net/url"
@@ -25,8 +25,12 @@ import (
 	"strings"
 	"sync"
 	"time"
+	"unicode"
 	"unicode/utf8"
 
+	"github.com/la5nta/wl2k-go/fbb"
+	"github.com/la5nta/wl2k-go/mailbox"
+
 	"github.com/la5nta/pat/cfg"
 	"github.com/la5nta/pat/internal/debug"
 	"github.com/la5nta/pat/internal/directories"
@@ -43,7 +47,8 @@ const (
 
 // Manager manages the forms subsystem
 type Manager struct {
-	config Config
+	config   Config
+	sequence Sequence
 
 	// postedFormData serves as an kv-store holding intermediate data for
 	// communicating form values submitted by the served HTML form files to
@@ -60,15 +65,26 @@ type Manager struct {
 	}
 }
 
+func (m *Manager) SeqSet(v int) error {
+	_, err := m.sequence.Set(int64(v))
+	return err
+}
+
+// LocatorProvider provides access to the current locator value
+type LocatorProvider interface {
+	Locator() string
+}
+
 // Config passes config options to the forms package
 type Config struct {
-	FormsPath  string
-	MyCall     string
-	Locator    string
-	AppVersion string
-	LineReader func() string
-	UserAgent  string
-	GPSd       cfg.GPSdConfig
+	FormsPath       string
+	SequencePath    string
+	SequenceFormat  string
+	MyCall          string
+	AppVersion      string
+	UserAgent       string
+	GPSd            cfg.GPSdConfig
+	LocatorProvider LocatorProvider
 }
 
 // FormFolder is a folder with forms. A tree structure with Form leaves and sub-Folder branches
@@ -93,12 +109,21 @@ var client = httpClient{http.Client{Timeout: 10 * time.Second}}
 func NewManager(conf Config) *Manager {
 	_ = os.MkdirAll(conf.FormsPath, 0o755)
 	retval := &Manager{
-		config: conf,
+		config:   conf,
+		sequence: OpenSequence(conf.SequencePath),
 	}
 	retval.postedFormData.m = make(map[string]Message)
 	return retval
 }
 
+func (m *Manager) Close() error {
+	if m == nil {
+		return nil
+	}
+	m.sequence.Close()
+	return nil
+}
+
 // GetFormsCatalogHandler reads all forms from config.FormsPath and writes them in the http response as a JSON object graph
 // This lets the frontend present a tree-like GUI for the user to select a form for composing a message
 func (m *Manager) GetFormsCatalogHandler(w http.ResponseWriter, r *http.Request) {
@@ -111,66 +136,107 @@ func (m *Manager) GetFormsCatalogHandler(w http.ResponseWriter, r *http.Request)
 	_ = json.NewEncoder(w).Encode(formFolder)
 }
 
-// PostFormDataHandler - When the user is done filling a form, the frontend posts the input fields to this handler,
-// which stores them in a map, so that other browser tabs can read the values back with GetFormDataHandler
-func (m *Manager) PostFormDataHandler(w http.ResponseWriter, r *http.Request) {
-	if err := r.ParseMultipartForm(10e6); err != nil {
-		http.Error(w, err.Error(), http.StatusBadRequest)
-		return
-	}
-	composeReply, _ := strconv.ParseBool(r.URL.Query().Get("composereply"))
-	templatePath := r.URL.Query().Get("template")
-	if templatePath == "" {
-		http.Error(w, "template query param missing", http.StatusBadRequest)
-		log.Printf("template query param missing %s %s", r.Method, r.URL.Path)
-		return
-	}
-	templatePath = m.abs(templatePath)
-	// Make sure we don't escape FormsPath
-	if !directories.IsInPath(m.config.FormsPath, templatePath) {
-		http.Error(w, fmt.Sprintf("%s escapes forms directory", templatePath), http.StatusForbidden)
-		return
-	}
+// PostFormDataHandler handles both HTML form submissions and text-only template submissions.
+// The handler detects the content type and processes accordingly, storing the results in
+// the forms map for retrieval by other browser tabs.
+func (m *Manager) PostFormDataHandler(mboxRoot string) http.HandlerFunc {
+	return func(w http.ResponseWriter, r *http.Request) {
+		inReplyTo := r.URL.Query().Get("in-reply-to")
+		templatePath := r.URL.Query().Get("template")
+		if templatePath == "" {
+			http.Error(w, "template query param missing", http.StatusBadRequest)
+			log.Printf("template query param missing %s %s", r.Method, r.URL.Path)
+			return
+		}
+		formInstanceKey, err := r.Cookie("forminstance")
+		if err != nil {
+			http.Error(w, err.Error(), http.StatusBadRequest)
+			log.Printf("missing cookie %s %s", templatePath, r.URL)
+			return
+		}
 
-	formInstanceKey, err := r.Cookie("forminstance")
-	if err != nil {
-		http.Error(w, err.Error(), http.StatusBadRequest)
-		log.Printf("missing cookie %s %s", templatePath, r.URL)
-		return
-	}
-	fields := make(map[string]string, len(r.PostForm))
-	for key, values := range r.PostForm {
-		fields[strings.TrimSpace(strings.ToLower(key))] = values[0]
-	}
+		var formValues, promptResponses map[string]string
+		switch contentType := r.Header.Get("Content-Type"); {
+		case strings.HasPrefix(contentType, "multipart/form-data"):
+			// Process HTML form submission
+			if err := r.ParseMultipartForm(10e6); err != nil {
+				http.Error(w, err.Error(), http.StatusBadRequest)
+				return
+			}
+			formValues = make(map[string]string, len(r.PostForm))
+			for key, values := range r.PostForm {
+				formValues[strings.TrimSpace(strings.ToLower(key))] = values[0]
+			}
+		case strings.HasPrefix(contentType, "application/json"):
+			// Process JSON template submission (from builtin text-only template editor)
+			var payload struct {
+				Responses map[string]string `json:"responses"`
+			}
+			if err := json.NewDecoder(r.Body).Decode(&payload); err != nil {
+				http.Error(w, err.Error(), http.StatusBadRequest)
+				return
+			}
+			formValues = map[string]string{"templateversion": m.getFormsVersion()}
+			promptResponses = payload.Responses
+		default:
+			http.Error(w, "unsupported content type", http.StatusBadRequest)
+			return
+		}
 
-	template, err := readTemplate(templatePath, formFilesFromPath(m.config.FormsPath))
-	switch {
-	case os.IsNotExist(err):
-		http.Error(w, err.Error(), http.StatusNotFound)
-		return
-	case err != nil:
-		http.Error(w, err.Error(), http.StatusInternalServerError)
-		log.Printf("failed to parse relevant form template (%q): %v", m.rel(templatePath), err)
-		return
-	}
-	msg, err := messageBuilder{
-		Template:    template,
-		FormValues:  fields,
-		Interactive: false,
-		IsReply:     composeReply,
-		FormsMgr:    m,
-	}.build()
-	if err != nil {
-		http.Error(w, err.Error(), http.StatusBadRequest)
-		log.Printf("%s %s: %s", r.Method, r.URL.Path, err)
-	}
+		templatePath = m.abs(templatePath)
+		// Make sure we don't escape FormsPath
+		if !directories.IsInPath(m.config.FormsPath, templatePath) {
+			http.Error(w, fmt.Sprintf("%s escapes forms directory", templatePath), http.StatusForbidden)
+			return
+		}
 
-	m.postedFormData.mu.Lock()
-	m.postedFormData.m[formInstanceKey.Value] = msg
-	m.postedFormData.mu.Unlock()
+		// Load template
+		template, err := readTemplate(m.abs(templatePath), formFilesFromPath(m.config.FormsPath))
+		switch {
+		case os.IsNotExist(err):
+			http.Error(w, err.Error(), http.StatusNotFound)
+			return
+		case err != nil:
+			http.Error(w, err.Error(), http.StatusInternalServerError)
+			log.Printf("failed to parse relevant form template (%q): %v", m.rel(templatePath), err)
+			return
+		}
 
-	m.cleanupOldFormData()
-	_, _ = io.WriteString(w, "<script>window.close()</script>")
+		// Load optional in-reply-to message
+		var inReplyToMsg *fbb.Message
+		if inReplyTo != "" {
+			var err error
+			inReplyToMsg, err = mailbox.OpenMessage(filepath.Join(mboxRoot, inReplyTo+mailbox.Ext))
+			if err != nil {
+				http.Error(w, err.Error(), http.StatusBadRequest)
+				log.Printf("failed to load in-reply-to (%q): %v", inReplyTo, err)
+				return
+			}
+		}
+
+		// Build message
+		msg, err := messageBuilder{
+			Template:        template,
+			FormValues:      formValues,
+			PromptResponses: promptResponses, // This is for text-only templates only
+			Interactive:     false,
+			InReplyToMsg:    inReplyToMsg,
+			FormsMgr:        m,
+		}.build()
+		if err != nil {
+			http.Error(w, err.Error(), http.StatusBadRequest)
+			log.Printf("%s %s: %s", r.Method, r.URL.Path, err)
+			return
+		}
+
+		// Save the result for the front-end to retrieve
+		m.postedFormData.mu.Lock()
+		m.postedFormData.m[formInstanceKey.Value] = msg
+		m.postedFormData.mu.Unlock()
+		m.cleanupOldFormData()
+
+		_, _ = io.WriteString(w, "<script>window.close()</script>")
+	}
 }
 
 // GetFormDataHandler is the counterpart to PostFormDataHandler. Returns the form field values to the frontend
@@ -197,6 +263,66 @@ func (m *Manager) GetPostedFormData(key string) (Message, bool) {
 	return v, ok
 }
 
+// GetFormAssetHandler serves files referenced by HTML forms (through {FormFolder}).
+//
+// It's primary use case is to load stylesheets and smilar resources.
+func (m *Manager) GetFormAssetHandler(w http.ResponseWriter, r *http.Request) {
+	path := m.abs(r.URL.Path)
+	// Make sure we don't escape FormsPath
+	if !directories.IsInPath(m.config.FormsPath, path) {
+		http.Error(w, fmt.Sprintf("%s escapes forms directory", path), http.StatusForbidden)
+		return
+	}
+	http.ServeFile(w, r, path)
+}
+
+// GetTemplateDataHandler serves partially processed template content.
+//
+// It's primary use case is to provide template files to the text-only template editor.
+func (m *Manager) GetTemplateDataHandler(mboxRoot string) http.HandlerFunc {
+	return func(w http.ResponseWriter, r *http.Request) {
+		templatePath := r.URL.Query().Get("template")
+		// Make sure we don't escape FormsPath
+		if !directories.IsInPath(m.config.FormsPath, m.abs(templatePath)) {
+			http.Error(w, fmt.Sprintf("%s escapes forms directory", templatePath), http.StatusForbidden)
+			return
+		}
+		template, err := readTemplate(m.abs(templatePath), formFilesFromPath(m.config.FormsPath))
+		if err != nil {
+			http.NotFound(w, r)
+			return
+		}
+		// Load optional in-reply-to message
+		var inReplyToMsg *fbb.Message
+		if inReplyTo := r.URL.Query().Get("in-reply-to"); inReplyTo != "" {
+			var err error
+			inReplyToMsg, err = mailbox.OpenMessage(filepath.Join(mboxRoot, inReplyTo+mailbox.Ext))
+			if err != nil {
+				http.Error(w, err.Error(), http.StatusBadRequest)
+				log.Printf("failed to load in-reply-to (%q): %v", inReplyTo, err)
+				return
+			}
+		}
+
+		// Open the template
+		f, err := os.Open(template.Path)
+		if err != nil {
+			http.Error(w, err.Error(), http.StatusInternalServerError)
+			return
+		}
+		defer f.Close()
+
+		// Process insertion tags (for preview)
+		var buf bytes.Buffer
+		if _, err := io.Copy(&buf, newTrimBomReader(f)); err != nil {
+			panic(err)
+		}
+		text := insertionTagReplacer(m, inReplyToMsg, templatePath, "<", ">")(buf.String())
+
+		io.WriteString(w, text)
+	}
+}
+
 // GetFormTemplateHandler serves a template's HTML form (filled-in with instance values)
 func (m *Manager) GetFormTemplateHandler(w http.ResponseWriter, r *http.Request) {
 	templatePath := r.URL.Query().Get("template")
@@ -224,11 +350,12 @@ func (m *Manager) GetFormTemplateHandler(w http.ResponseWriter, r *http.Request)
 	}
 	formPath := template.InputFormPath
 	if formPath == "" {
-		http.Error(w, "requested template does not provide a HTML form", http.StatusNotFound)
+		// This is a text-only template. Redirect to template editor.
+		http.Redirect(w, r, "/ui/template?"+r.URL.Query().Encode(), http.StatusFound)
 		return
 	}
 
-	responseText, err := m.fillFormTemplate(formPath, "/api/form?"+r.URL.Query().Encode(), nil)
+	responseText, err := m.fillFormTemplate(formPath, nil, "/api/form?"+r.URL.Query().Encode(), nil)
 	if err != nil {
 		http.Error(w, err.Error(), http.StatusBadRequest)
 		log.Printf("problem filling form template file %s %s: can't open template %s. Err: %s", r.Method, r.URL.Path, formPath, err)
@@ -308,7 +435,7 @@ func (m *Manager) downloadAndUnzipForms(ctx context.Context, downloadLink string
 		return fmt.Errorf("can't download update ZIP: %w", err)
 	}
 	defer resp.Body.Close()
-	f, err := ioutil.TempFile(os.TempDir(), "pat")
+	f, err := os.CreateTemp(os.TempDir(), "pat")
 	if err != nil {
 		return fmt.Errorf("can't create temp file for download: %w", err)
 	}
@@ -325,7 +452,7 @@ func (m *Manager) downloadAndUnzipForms(ctx context.Context, downloadLink string
 }
 
 // RenderForm finds the associated form and returns the filled-in form in HTML given the contents of a form attachment
-func (m *Manager) RenderForm(data []byte, composeReply bool) (string, error) {
+func (m *Manager) RenderForm(data []byte, inReplyToMsg *fbb.Message, inReplyToPath string) (string, error) {
 	type Node struct {
 		XMLName xml.Name
 		Content []byte `xml:",innerxml"`
@@ -363,7 +490,7 @@ func (m *Manager) RenderForm(data []byte, composeReply bool) (string, error) {
 
 	filesMap := formFilesFromPath(m.config.FormsPath)
 	switch {
-	case composeReply:
+	case inReplyToPath != "":
 		replyTemplate := formParams["reply_template"]
 		if replyTemplate == "" {
 			return "", errors.New("missing reply_template tag in form XML for a reply message")
@@ -379,8 +506,8 @@ func (m *Manager) RenderForm(data []byte, composeReply bool) (string, error) {
 		if err != nil {
 			return "", fmt.Errorf("failed to read referenced reply template: %w", err)
 		}
-		submitURL := "/api/form?composereply=true&template=" + url.QueryEscape(m.rel(template.Path))
-		return m.fillFormTemplate(template.InputFormPath, submitURL, formVars)
+		submitURL := "/api/form?in-reply-to=" + url.QueryEscape(inReplyToPath) + "&template=" + url.QueryEscape(m.rel(template.Path))
+		return m.fillFormTemplate(template.InputFormPath, inReplyToMsg, submitURL, formVars)
 	default:
 		displayForm := formParams["display_form"]
 		if displayForm == "" {
@@ -394,16 +521,20 @@ func (m *Manager) RenderForm(data []byte, composeReply bool) (string, error) {
 		if path == "" {
 			return "", fmt.Errorf("display from not found: %q", displayForm)
 		}
-		return m.fillFormTemplate(path, "", formVars)
+		return m.fillFormTemplate(path, inReplyToMsg, "", formVars)
 	}
 }
 
 // ComposeTemplate composes a message from a template (templatePath) by prompting the user through stdio.
 //
 // It combines all data needed for the whole template-based message: subject, body, and attachments.
-func (m *Manager) ComposeTemplate(templatePath string, subject string) (Message, error) {
+func (m *Manager) ComposeTemplate(templatePath string, subject string, inReplyToMsg *fbb.Message, lineReader func() string) (Message, error) {
 	template, err := readTemplate(templatePath, formFilesFromPath(m.config.FormsPath))
-	if err != nil {
+	switch {
+	case os.IsNotExist(err) && !filepath.IsAbs(templatePath):
+		// Try resolving the path relative to forms directory.
+		return m.ComposeTemplate(m.abs(templatePath), subject, inReplyToMsg, lineReader)
+	case err != nil:
 		return Message{}, err
 	}
 
@@ -413,10 +544,13 @@ func (m *Manager) ComposeTemplate(templatePath string, subject string) (Message,
 	}
 	fmt.Printf("Form '%s', version: %s\n", m.rel(template.Path), formValues["templateversion"])
 	return messageBuilder{
-		Template:    template,
-		FormValues:  formValues,
 		Interactive: true,
-		FormsMgr:    m,
+		LineReader:  lineReader,
+
+		Template:     template,
+		FormValues:   formValues,
+		FormsMgr:     m,
+		InReplyToMsg: inReplyToMsg,
 	}.build()
 }
 
@@ -446,7 +580,7 @@ func (m *Manager) innerRecursiveBuildFormFolder(rootPath string, filesMap formFi
 			}
 			folder.Folders = append(folder.Folders, subfolder)
 			folder.FormCount += subfolder.FormCount
-			return nil
+			return fs.SkipDir
 		case !strings.EqualFold(filepath.Ext(d.Name()), txtFileExt):
 			return nil
 		default:
@@ -457,10 +591,6 @@ func (m *Manager) innerRecursiveBuildFormFolder(rootPath string, filesMap formFi
 			}
 			// Relative paths for the JSON response
 			template.Path = m.rel(template.Path)
-			// Ignore templates without a input form (we don't support text-only templates in the web gui yet)
-			if template.InputFormPath == "" {
-				return nil
-			}
 			folder.Forms = append(folder.Forms, template)
 			folder.FormCount++
 			return nil
@@ -524,7 +654,7 @@ func (m *Manager) gpsPos() (gpsd.Position, error) {
 	return conn.NextPosTimeout(3 * time.Second)
 }
 
-func (m *Manager) fillFormTemplate(templatePath string, formDestURL string, formVars map[string]string) (string, error) {
+func (m *Manager) fillFormTemplate(templatePath string, inReplyToMsg *fbb.Message, formDestURL string, formVars map[string]string) (string, error) {
 	data, err := readFile(templatePath)
 	if err != nil {
 		return "", err
@@ -535,7 +665,7 @@ func (m *Manager) fillFormTemplate(templatePath string, formDestURL string, form
 	data = strings.ReplaceAll(data, "http://localhost:8001", formDestURL) // Some Canada BC forms are hardcoded to this URL
 
 	// Substitute insertion tags and variables
-	data = insertionTagReplacer(m, "{", "}")(data)
+	data = insertionTagReplacer(m, inReplyToMsg, templatePath, "{", "}")(data)
 	data = variableReplacer("{", "}", formVars)(data)
 
 	return data, nil
@@ -547,7 +677,15 @@ func (m *Manager) getFormsVersion() string {
 		debug.Printf("failed to open version file: %v", err)
 		return "unknown"
 	}
-	return strings.TrimSpace(str)
+	// Drop any whitespace in the string
+	// (version 1.1.6.0 was released as `1.1.6\t.0`).
+	str = strings.Map(func(r rune) rune {
+		if unicode.IsSpace(r) {
+			return -1
+		}
+		return r
+	}, str)
+	return str
 }
 
 func (m *Manager) cleanupOldFormData() {
